# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class Template2LevelSAEMCMC(Template):       
    'Fits 2-level random intercept Normal model using MCMC to a sample dataset and uses this with a population dataset to create SAE estimates.'

    __version__ = '1.0.0'

    tags = [ 'Model', '2-Level', 'SAE']
    engines = ['CustomC']

    inputs = '''
y = DataVector('Response: ')
D = Text('Specify distribution: ', ['Normal', 'Binomial', 'Poisson', '-ve Binomial'])
if D == 'Binomial':
    n = DataVector('Denominator: ',  help='e.g. if modelling a binary 0/1 response, select a constant of ones.')
    link = Text('Specify link function: ', ['logit', 'probit', 'cloglog'])
if D == 'Poisson':
    link = Text(value = 'ln')
    offset = Boolean('Is there an offset: ', help="An offset allows you to model rates, instead of raw counts.")
    if offset:
        n = DataVector('Offset: ')
if D == '-ve Binomial':
    offset = Boolean('Is there an offset: ')
    if offset:
        n = DataVector('Offset: ')
if D == 'Normal':
    transform = Text('Transformation:', ['Box-Cox', 'Log', 'None'])
    shiftval = Text(value='')
    if transform == 'Box-Cox':
        lambtype = Text('Lambda:', ['Automatic', 'Modelled', 'Manual'])
        if lambtype == 'Manual':
            lambval = Text('Value')
        else:
            lambval = Text(value='')
else:
    transform = Text(value='None')
L2ID = IDVector('Level 2 ID: ')
x = DataMatrix('Explanatory variables: ', allow_cat = True)

povind = Boolean('Do you want to calculate poverty related estimates, e.g. Head count ratio?: ')
if povind:
    userthresh = Text('Threshold value: ', ['Automatic', 'Manual'])
    if userthresh == 'Manual':
        thresh = Text('Threshold value')
    else:
        thresh = Text(value='')
ineqind = Boolean('Do you want to calculate inequality related estimates, e.g. GINI index?: ')

numproc = Integer('Number of parallel cores: ', default=6)

popdata = DataFileChooser('Population dataset')
popid = IDVector(value=str(L2ID), data_src=popdata)
x1 = DataVector(value=list(x)[0], data_src=popdata) # Used to determine length of population dataset for prediction parameter
popx = DataMatrix(value=list(x.orignames), data_src=popdata, allow_cat = True)
if D == 'Binomial':
    popn = DataVector(value=str(n), data_src=popdata)
if D == 'Poisson' or D == '-ve Binomial':
    if offset:
        popn = DataVector(value=str(n), data_src=popdata)

matchvar = DataVector(value="_idmatch", data_src=popdata)

yj_mean = ParamVector(parents=[popid], as_scalar=False)
yj_std = ParamVector(parents=[popid], as_scalar=False)

yj_q10 = ParamVector(parents=[popid], as_scalar=False)
yj_q20 = ParamVector(parents=[popid], as_scalar=False)
yj_q25 = ParamVector(parents=[popid], as_scalar=False)
yj_q50 = ParamVector(parents=[popid], as_scalar=False)
yj_q75 = ParamVector(parents=[popid], as_scalar=False)
yj_q80 = ParamVector(parents=[popid], as_scalar=False)
yj_q90 = ParamVector(parents=[popid], as_scalar=False)
if povind:
    yj_headcount = ParamVector(parents=[popid], as_scalar=False)
    yj_povgap = ParamVector(parents=[popid], as_scalar=False)
if ineqind:
    yj_gini = ParamVector(parents=[popid], as_scalar=False)
    yj_qshare = ParamVector(parents=[popid], as_scalar=False)

y_pred = ParamVector(parents=[x1], as_scalar=False, monitor=False, modelled=False)

beta = ParamVector(parents=[x], as_scalar=False)
u = ParamVector(parents=[L2ID], as_scalar=False)
e = ParamVector(parents=[y], as_scalar=False, monitor=False, modelled=False)

if D == 'Normal':
    tau = ParamScalar()
    sigma2 = ParamScalar(modelled = False)
if D == '-ve Binomial':
    alpha = ParamScalar()
    rho = ParamVector(parents = [y])

tau_u = ParamScalar()
sigma_u = ParamScalar(modelled = False)

if transform == 'Box-Cox':
    if lambtype == 'Modelled':
        lamb = ParamScalar()
        jac = ParamScalar()
    else:
        lamb = ParamScalar(modelled = False)
        jac = ParamScalar(modelled = False)
    shift = ParamScalar(modelled = False)
if transform == 'Log':
    shift = ParamScalar(modelled = False)

deviance = ParamScalar(modelled = False)
'''

    latex = r'''
\begin{aligned} 
% if D == 'Normal':
% if transform == 'None':
 \mbox{${y}}_{i}
% endif
% if transform == 'Log':
 \mbox{log(${y}}_{i}} - ${shiftval}\mbox{)}
% endif
% if transform == 'Box-Cox':
% if lambtype == 'Modelled':
 \mbox{box-cox(${y}}_{i} - ${shiftval}\mbox{, }\lambda\mbox{)}
% else:
 \mbox{box-cox(${y}}_{i} - ${shiftval}\mbox{, ${lambval})}
% endif
% endif
 & \sim \mbox{N}(\mu_{i}, \sigma^2) \\ 
\mu_{i} & =
% endif
% if D == 'Binomial':
 \mbox{${y}}_{i} & \sim \mbox{Binomial}(\mbox{${n}}_{i}, \pi_{i}) \\ 
\mbox{${link}}(\pi_{i}) & =
% endif
% if D == 'Poisson':
 \mbox{${y}}_{i} & \sim \mbox{Poisson}(\pi_{i}) \\ 
\mbox{${link}}(\pi_{i}) & =
% if offset:
\mbox{${n}}_{i} +
% endif
% endif
${mmulttex(x, r'\beta', 'i')} + u_{\mbox{${L2ID}}[i]}\\ 
u_{\mbox{${L2ID}}[i]} & \sim \mbox{N}(0,\sigma^2_u) \\ 

% if len(x) > 2:
\beta_{0}, \dots, \beta_{${len(x)-1}} & \propto 1 \\ 
% elif len(x) == 2:
\beta_{0}, \beta_{${len(x)-1}} & \propto 1 \\ 
% else:
\beta_{0} & \propto 1 \\ 
% endif
% if D == 'Normal':
\tau & \sim \Gamma (0.001,0.001) \\ 
\sigma^2 & = 1 / \tau \\ 
% endif
\tau_u & \sim \Gamma (0.001,0.001) \\ 
\sigma_u^2 & = 1 / \tau_u
\end{aligned} 
'''

    customccode = '''
	{
% if D == 'Normal':
% if transform == 'Box-Cox' and lambtype == 'Modelled':
		jac = 0.0;
		for (int i = 0; i < mat_y.rows(); ++i) {
			jac += std::log(mat_y(i, 0)); 
		}
		jac = std::exp(jac/mat_y.rows());
		jac = std::pow(jac, lamb-1);
% endif

		// Multivariate step for beta
		for (int i = 0; i < mat_u.rows(); i++) {
			double sumy = 0.0;
			for (int id_k = 0; id_k < idmapsizes_${L2ID}[i]; id_k++) {
				int k = idmap_${L2ID}[i][id_k];
% if transform == 'Box-Cox' and lambtype == 'Modelled':
				sumy += ((std::pow(mat_y(k, 0), lamb) - 1) / (lamb*jac));
% else:
				sumy += mat_y(k, 0);
% endif
			}
			double mean = sumy;
			for (int j = 0; j < mat_beta.rows(); j++) {
				double sumx = 0.0;
				for (int id_k = 0; id_k < idmapsizes_${L2ID}[i]; id_k++) {
					int k = idmap_${L2ID}[i][id_k];
					sumx += mat_x(k, j);
				}
				mean -= mat_beta(j, 0) * sumx;
			}

			double du = tau_u + tau * idmapsizes_${L2ID}[i];
			mat_u(i, 0) = dnorm((tau * mean) / du, du);
		}

% if transform == 'Box-Cox' and lambtype == 'Modelled':
		RectMatrix ydiff(mat_y.rows(), 1);
		for (int i = 0; i < ydiff.rows(); ++i) {
			ydiff(i, 0) = ((std::pow(mat_y(i, 0), lamb) - 1) / (lamb*jac)) - mat_u(int(${L2ID}[i]), 0);
		}
% else:
		RectMatrix ydiff(mat_y);
		for (int i = 0; i < ydiff.rows(); ++i) {
			ydiff(i, 0) -= mat_u(int(${L2ID}[i]), 0);
		}
% endif

		// Setting up constant terms for beta step
		bool fail = false;
		RectMatrix xtxchol = matrix_cholesky(mat_x.T() * mat_x, false, fail);
		RectMatrix mean = matrix_cholsolve(xtxchol, mat_x.T() * ydiff);

		// Multivariate step for beta (Model of interest)
		DiagMatrix taudiag = identity(${len(x)}) / tau;
		SymMatrix variance = matrix_cholsolve(xtxchol, taudiag);
		mat_beta = dmultnormal(mean, variance, fail);

		// Note: ydiff is y - u from above
		mat_e = ydiff - mat_x * mat_beta;

		tau = dgamma(0.001 + 0.5 * mat_y.rows(), 0.001 + 0.5 * (mat_e.T() * mat_e)(0,0));
		sigma2 = 1/tau;

% if transform == 'Box-Cox' and lambtype == 'Modelled':
		// Step for lambda
		double oldlamb = lamb;
		double oldjac = jac;

		double llcurr = -0.5 * tau * (mat_e.T() * mat_e)(0,0);

		do {
			lamb = dnorm(lamb, 1.0/lamb_sd);
		} while (lamb == 0);

		jac = 0.0;
		for (int i = 0; i < mat_y.rows(); ++i) {
			jac += std::log(mat_y(i, 0)); 
		}
		jac = std::exp(jac/mat_y.rows());
		jac = std::pow(jac, lamb-1);

		for (size_t i = 0; i < mat_e.rows(); ++i) {
			mat_e(i, 0) -= ((std::pow(mat_y(i, 0), oldlamb) - 1) / (oldlamb*oldjac));
			mat_e(i, 0) += ((std::pow(mat_y(i, 0), lamb) - 1) / (lamb*jac));
		}

		double llnew = -0.5 * tau * (mat_e.T() * mat_e)(0,0);

		int change = 1;
		if (isFinite(llnew) == 0) {
			change = 0;
		} else {
			if (llnew > llcurr) {
				;
			} else {
				double rand = random01();
				if (rand > exp(llnew - llcurr))
					change = 0;
			}
		}
		if (change == 0) {
			for (size_t i = 0; i < mat_e.rows(); ++i) {
				mat_e(i, 0) -= ((std::pow(mat_y(i, 0), lamb) - 1) / (lamb*jac));
				mat_e(i, 0) += ((std::pow(mat_y(i, 0), oldlamb) - 1) / (oldlamb*oldjac));
			}
			lamb = oldlamb;
			jac = oldjac;
		} else {
			lamb_accept++;
		}
% endif

		deviance = (tau*(mat_e.T() * mat_e)(0,0)) + (log(3.14159265)-log(tau))*length(${y}) + (2.0*0.346573590279973*length(${y})); // Last term is log(sqrt(2))
% else:
		for (int i = 0; i < mat_x.cols(); ++i) {
			double llcurr = 0.0;
			{
				for (int j = 0; j < mat_x.rows(); ++j) {
					double xb = 0;
					for (int k = 0; k < mat_x.cols(); ++k) {
						xb += mat_x(j, k) * mat_beta(k, 0);
					}
					xb += mat_u(int(${L2ID}[j]), 0);
% if D == 'Binomial':
% if link == 'logit':
					double temp = -log1p(exp(-xb));
					llcurr += (mat_n(j, 0) - mat_y(j, 0)) * (-xb + temp) + mat_y(j, 0) * temp;
% endif
% if link == 'probit':
					double temp = normprob(-xb);
					llcurr += (mat_n(j, 0) - mat_y(j, 0)) * log(temp) + mat_y(j, 0) * log(1.0 - temp);
% endif
% if link == 'cloglog':
					double temp = -exp(xb);
					llcurr += (mat_n(j, 0) - mat_y(j, 0)) * temp + mat_y(j, 0) * log(1.0 - exp(temp));
% endif
% endif
% if D == 'Poisson':
% if offset:
					xb += mat_n(j, 0);
% endif
					llcurr += mat_y(j, 0) * xb - exp(xb);
% endif
% if D == '-ve Binomial':
% if offset:
					xb += mat_n(j, 0);
% endif
					llcurr += mat_y(j, 0) * xb - exp(xb) * mat_rho(j, 0);
% endif
				}
			}
			double old = mat_beta(i, 0);
			mat_beta(i, 0) = dnorm(mat_beta(i, 0), 1.0/beta_sd[i]);

			double llnew = 0.0;
			{
				for (int j = 0; j < mat_x.rows(); ++j) {
					double xb = 0;
					for (int k = 0; k < mat_x.cols(); ++k) {
						xb += mat_x(j, k) * mat_beta(k, 0);
					}
					xb += mat_u(int(${L2ID}[j]), 0);
% if D == 'Binomial':
% if link == 'logit':
					double temp = -log1p(exp(-xb));
					llnew += (mat_n(j, 0) - mat_y(j, 0)) * (-xb + temp) + mat_y(j, 0) * temp;
% endif
% if link == 'probit':
					double temp = normprob(-xb);
					llnew += (mat_n(j, 0) - mat_y(j, 0)) * log(temp) + mat_y(j, 0) * log(1.0 - temp);
% endif
% if link == 'cloglog':
					double temp = exp(-exp(xb));
					llnew += (mat_n(j, 0) - mat_y(j, 0)) * log(temp) + mat_y(j, 0) * log(1.0 - temp);
% endif
% endif
% if D == 'Poisson':
% if offset:
					xb += mat_n(j, 0);
% endif
					llnew += mat_y(j, 0) * xb - exp(xb);
% endif
% if D == '-ve Binomial':
% if offset:
					xb += mat_n(j, 0);
% endif
					llnew += mat_y(j, 0) * xb - exp(xb) * mat_rho(j, 0);
% endif
				}
			}

			int change = 1;
			if (isFinite(llnew) == 0) {
				change = 0;
			} else {
				if (llnew > llcurr) {
					;
				} else {
					double rand = random01();
					if (rand > exp(llnew - llcurr))
						change = 0;
				}
			}
			if (change == 0) {
				mat_beta(i, 0) = old;
			} else {
				beta_accept[i]++;
			}
		}

		for (int i = 0; i < mat_u.rows(); ++i) {
			double llcurr = 0.0;
			for (int id_j = 0; id_j < length(idmap_${L2ID}[i]); id_j++) {
				unsigned int j = idmap_${L2ID}[i][id_j];
				double xb = 0;
				for (int k = 0; k < mat_x.cols(); ++k) {
					xb += mat_x(j, k) * mat_beta(k, 0);
				}
				xb += mat_u(i, 0);
% if D == 'Binomial':
% if link == 'logit':
				double temp = -log1p(exp(-xb));
				llcurr += (mat_n(j, 0) - mat_y(j, 0)) * (-xb + temp) + mat_y(j, 0) * temp;
% endif
% if link == 'probit':
				double temp = normprob(-xb);
				llcurr += (mat_n(j, 0) - mat_y(j, 0)) * log(temp) + mat_y(j, 0) * log(1.0 - temp);
% endif
% if link == 'cloglog':
				double temp = -exp(xb);
				llcurr += (mat_n(j, 0) - mat_y(j, 0)) * temp + mat_y(j, 0) * log(1.0 - exp(temp));
% endif
% endif
% if D == 'Poisson':
% if offset:
				xb += mat_n(j, 0);
% endif
				llcurr += mat_y(j, 0) * xb - exp(xb);
% endif
% if D == '-ve Binomial':
% if offset:
				xb += mat_n(j, 0);
% endif
				llcurr += mat_y(j, 0) * xb - exp(xb) * mat_rho(j, 0);
% endif
			}
			llcurr -= 0.5 * tau_u * pow(mat_u(i, 0), 2);

			double old = mat_u(i, 0);
			mat_u(i, 0) = dnorm(mat_u(i, 0), 1.0/u_sd[i]);

			double llnew = 0.0;
			for (int id_j = 0; id_j < length(idmap_${L2ID}[i]); id_j++) {
				unsigned int j = idmap_${L2ID}[i][id_j];
				double xb = 0;
				for (int k = 0; k < mat_x.cols(); ++k) {
					xb += mat_x(j, k) * mat_beta(k, 0);
				}
				xb += mat_u(i, 0);
% if D == 'Binomial':
% if link == 'logit':
				double temp = -log1p(exp(-xb));
				llnew += (mat_n(j, 0) - mat_y(j, 0)) * (-xb + temp) + mat_y(j, 0) * temp;
% endif
% if link == 'probit':
				double temp = normprob(-xb);
				llnew += (mat_n(j, 0) - mat_y(j, 0)) * log(temp) + mat_y(j, 0) * log(1.0 - temp);
% endif
% if link == 'cloglog':
				double temp = -exp(xb);
				llnew += (mat_n(j, 0) - mat_y(j, 0)) * temp + mat_y(j, 0) * log(1.0 - exp(temp));
% endif
% endif
% if D == 'Poisson':
% if offset:
				xb += mat_n(j, 0);
% endif
				llnew += mat_y(j, 0) * xb - exp(xb);
% endif
% if D == '-ve Binomial':
% if offset:
				xb += mat_n(j, 0);
% endif
				llnew += mat_y(j, 0) * xb - exp(xb) * mat_rho(j, 0);
% endif
			}
			llnew -= 0.5 * tau_u * pow(mat_u(i, 0), 2);

			bool change = true;
			if (isFinite(llnew) == 0) {
				change = false;
			} else {
				if (llnew > llcurr) {
					;
				} else {
					double rand = random01();
					if (rand > exp(llnew - llcurr))
						change = 0;
				}
			}
			if (change == false) {
				mat_u(i, 0) = old;
			} else {
				u_accept[i]++;
			}
		}

% if D == '-ve Binomial':
		{
			double sum = 0.0;
			for (int j = 0; j < mat_xint.rows(); ++j) {
				sum += mat_rho(j, 0) - log(mat_rho(j, 0));
			}

			double llcurr = (((-sum * alpha) + (mat_x.rows() * alpha * log(alpha)) - (0.0005 * pow(log(alpha), 2)) - log(alpha)) - mat_x.rows() * lgamma(alpha));

			double old = alpha;
			alpha = dnorm(alpha, 1.0/alpha_sd);

			double llnew = (((-sum * alpha) + (mat_x.rows() * alpha * log(alpha)) - (0.0005 * pow(log(alpha), 2)) - log(alpha)) - mat_x.rows() * lgamma(alpha));

			bool change = true;
			if (isFinite(llnew) == 0) {
				change = false;
			} else {
				if (llnew > llcurr) {
					;
				} else {
					double rand = random01();
					if (rand > exp(llnew - llcurr))
						change = 0;
				}
			}
			if (change == false) {
				alpha = old;
			} else {
				alpha_accept++;
			}

			for (int i = 0; i < mat_rho.rows(); ++i) {
				double xb = 0;
				for (int k = 0; k < mat_xint.cols(); ++k) {
					xb += mat_xint(i, k) * mat_beta(k, 0);
				}
				mat_rho(i, 0) = dgamma(alpha + mat_y(i, 0), alpha + exp(xb));
			}
		}
% endif

		deviance = 0.0;
		for (int j = 0; j < mat_x.rows(); ++j) {
			double xb = 0;
			for (int k = 0; k < mat_x.cols(); ++k) {
				xb += mat_x(j, k) * mat_beta(k, 0);
			}
			xb += mat_u(int(${L2ID}[j]), 0);
% if D == 'Binomial':
% if link == 'logit':
			double temp = -log1p(exp(-xb));
			mat_e(j, 0) = mat_y(j, 0) - (1.0 / 1.0 + exp(-xb)) * mat_n(j, 0);
			deviance += (mat_n(j, 0) - mat_y(j, 0)) * (-xb + temp) + mat_y(j, 0) * temp;
% endif
% if link == 'probit':
			double temp = normprob(-xb);
			mat_e(j, 0) = mat_y(j, 0) - normprob(xb) * mat_n(j, 0);
			deviance += (mat_n(j, 0) - mat_y(j, 0)) * log(temp) + mat_y(j, 0) * log(1.0 - temp);
% endif
% if link == 'cloglog':
			double temp = -exp(xb);
			mat_e(j, 0) = mat_y(j, 0) - exp(temp) * mat_n(j, 0);
			deviance += (mat_n(j, 0) - mat_y(j, 0)) * temp + mat_y(j, 0) * log(1.0 - exp(temp));
% endif
			deviance += logfact(mat_n(j, 0)) - logfact(mat_y(j, 0)) - logfact(mat_n(j, 0) - mat_y(j, 0));
% endif
% if D == 'Poisson':
% if offset:
			xb += mat_n(j, 0);
% endif
			mat_e(j, 0) = mat_y(j, 0) - exp(xb);
			deviance += mat_y(j, 0) * xb - exp(xb);
			deviance += logfact(mat_y(j, 0));
% endif
% if D == '-ve Binomial':
% if offset:
			xb += mat_n(j, 0);
% endif
			mat_e(j, 0) = mat_y(j, 0) - exp(xb);
			deviance += mat_y(j, 0) * xb - exp(xb) * mat_rho(j, 0);
			deviance += logfact(mat_y(j, 0));
% endif
		}
		deviance *= -2;
% endif
		tau_u = dgamma(0.001 + 0.5 * mat_u.rows(), 0.001 + 0.5 * (mat_u.T() * mat_u)(0, 0));
		sigma_u = 1/tau_u;

		static std::unordered_map<std::string, std::vector<std::future<void> > > proclists;
		proclists.insert(std::make_pair(rngid, std::vector<std::future<void> >(${numproc})));
		auto &procs = proclists[rngid];

		static std::unordered_map<std::string, bool> modeldone; /* Flag to indicate if model estimation has ended */
		if (runstate == 3 && iter == numiter - 1) {
			modeldone[rngid] = true;
		} else {
			modeldone[rngid] = false;
		}
		bool &done = modeldone[rngid];

		static std::unordered_map<std::string, std::vector<double> > y_Ms;
		static std::unordered_map<std::string, std::vector<double> > y_Ss;
		static std::unordered_map<std::string, std::vector<size_t> > zerocounts;
		static std::unordered_map<std::string, std::vector<size_t> > zerocountsgrp;

		if (runstate == 0) {
			y_Ms.try_emplace(rngid, std::vector<double>());
			y_Ss.try_emplace(rngid, std::vector<double>());
			zerocounts.try_emplace(rngid, std::vector<size_t>(mat_popx.rows()));
			zerocountsgrp.try_emplace(rngid, std::vector<size_t>(mat_x.rows()));
		}

		if (runstate == 3 && iter == 0) {
			auto &y_M = y_Ms[rngid];
			y_M.assign(y_pred_M, y_pred_M + mat_popx.rows());
			auto &y_S = y_Ss[rngid];
			y_S.assign(y_pred_S, y_pred_S + mat_popx.rows());

			auto &zerocount = zerocounts[rngid];
			auto &zerocountgrp = zerocountsgrp[rngid];

			// WARNING: Indexing the stored chains here relies on implementation details which may change.
			auto chunksize = idmap_${popdata}_${L2ID}.size() / ${numproc};
			auto endchunksize = idmap_${popdata}_${L2ID}.size() - (chunksize * (${numproc} - 1));
			for (int p = 0; p < ${numproc}; ++p) {
				procs[p] = std::async(std::launch::async, 
					[ /* Captured variables */
						rngid,
						runstate,
						seed,
						idmapsizes_${popdata}_${L2ID}, /* Input, Constant across iterations */
						idmap_${popdata}_${L2ID}, /* Input, Constant across iterations */
						yj_mean_chain, /* Output, Varies across iterations */
						&mat_yj_mean, /* Output, Varies across iterations (should only be used for last iteration) */
						yj_std_chain, /* Output, Varies across iterations */
						&mat_yj_std, /* Output, Varies across iterations (should only be used for last iteration) */
						yj_q10_chain, /* Output, Varies across iterations */
						&mat_yj_q10, /* Output, Varies across iterations (should only be used for last iteration) */
						yj_q20_chain, /* Output, Varies across iterations */
						&mat_yj_q20, /* Output, Varies across iterations (should only be used for last iteration) */
						yj_q25_chain, /* Output, Varies across iterations */
						&mat_yj_q25, /* Output, Varies across iterations (should only be used for last iteration) */
						yj_q50_chain, /* Output, Varies across iterations */
						&mat_yj_q50, /* Output, Varies across iterations (should only be used for last iteration) */
						yj_q75_chain, /* Output, Varies across iterations */
						&mat_yj_q75, /* Output, Varies across iterations (should only be used for last iteration) */
						yj_q80_chain, /* Output, Varies across iterations */
						&mat_yj_q80, /* Output, Varies across iterations (should only be used for last iteration) */
						yj_q90_chain, /* Output, Varies across iterations */
						&mat_yj_q90, /* Output, Varies across iterations (should only be used for last iteration) */
% if povind:
						yj_headcount_chain, /* Output, Varies across iterations */
						&mat_yj_headcount, /* Output, Varies across iterations (should only be used for last iteration) */
						yj_povgap_chain, /* Output, Varies across iterations */
						&mat_yj_povgap, /* Output, Varies across iterations (should only be used for last iteration) */
% endif
% if ineqind:
						yj_gini_chain, /* Output, Varies across iterations */
						&mat_yj_gini, /* Output, Varies across iterations (should only be used for last iteration) */
						yj_qshare_chain, /* Output, Varies across iterations */
						&mat_yj_qshare, /* Output, Varies across iterations (should only be used for last iteration) */
% endif
						&y_pred,
						&y_M,
						&y_S,
						&zerocount,
						&zerocountgrp,
						${popdata}_${matchvar},  /* Input, Constant across iterations */
						numiter, /* Input, Constant across iterations */
						&iter, /* Input, Varies across iterations */
						start_iteration, /* Input, Constant across iterations */
						thinning, /* Input, Constant across iterations */
						&done, /* Input, Varies across iterations */
						beta_chain, /* Input, Varies across iterations */
						&mat_beta, /* Input, Varies across iterations (should only be used for last iteration) */
						&mat_popx,  /* Input, Constant across iterations */
						u_chain, /* Input, Constant across iterations */
						&mat_u, /* Input, Varies across iterations (should only be used for last iteration) */
% if D == 'Normal':
						tau_chain, /* Input, Constant across iterations */
						&tau,  /* Input, Varies across iterations (should only be used for last iteration) */
% endif
% if D == 'Binomial' or ((D == 'Poisson' or D == '-ve Binomial') and offset):
						${popdata}_${n},
% endif
% if D == '-ve Binomial':
						alpha_chain, /* Input, Constant across iterations */
						&alpha, /* Input, Varies across iterations */
						rho_chain, /* Input, Constant across iterations */
						&mat_rho, /* Input, Varies across iterations */
% endif
% if transform == 'Box-Cox' or transform == 'Log':
						shift,
% endif
% if transform == 'Box-Cox':
						lamb_chain,
						&lamb,
						jac_chain,
						&jac,
% endif
						tau_u_chain, /* Input, Constant across iterations */
						&tau_u /* Input, Varies across iterations (should only be used for last iteration) */
					]
					(size_t startidx, size_t endidx) { /* Function body */
						//SimpleRandomGenerator rng1 = SimpleRandomGenerator(); /* "rng" is hardcoded into the random number function, so we have to use this here even though it masks the outer definition */
						rngstate.try_emplace(rngid + std::to_string(startidx), SimpleRandomGenerator());
						auto &rng = rngstate[rngid + std::to_string(startidx)];
						rng.set_seed(seed + startidx);
						for (auto it = 0; it < numiter; it += thinning) {
							size_t iterind = start_iteration + floor(iter / thinning);
							size_t predind = start_iteration + (it / thinning);
							while (predind >= iterind && done == false) { /* Stay at least an iteration behind except for last iteration */
								std::cout << "Waiting for data to become available for current iteration (" << predind * thinning << ")..." << std::endl;
								std::this_thread::yield();
								std::this_thread::sleep_for(std::chrono::milliseconds(500));
								iterind = start_iteration + floor(iter / thinning); /* iter should have changed, so recalculate */
							};

							for (auto i = startidx; i < endidx; ++i) {
								//double sumy = 0.0;
								std::vector<double> ys(idmapsizes_${popdata}_${L2ID}[i]);
								for (int id_k = 0; id_k < idmapsizes_${popdata}_${L2ID}[i]; id_k++) {
									int k = idmap_${popdata}_${L2ID}[i][id_k];
									double mean = 0;
									if (it != numiter - thinning) {
										for (size_t j = 0; j < mat_popx.cols(); ++j) {
											mean += beta_chain[(predind * mat_beta.rows()) + j] * mat_popx(k, j);
										}
									} else { /* Values have not yet been written to the chain, so fetch from local variable */
										for (size_t j = 0; j < mat_popx.cols(); ++j) {
											mean += mat_beta(j, 0) * mat_popx(k, j);
										}
									}

									auto urow = ${popdata}_${matchvar}[k];

									if (urow != -1) {
										if (it != numiter - thinning) {
											size_t uind = (predind * mat_u.rows()) + urow;
											mean += u_chain[uind];
										} else { /* Values have not yet been written to the chain, so fetch from local variable */
											mean += mat_u(urow, 0);
										}
									} else {
										if (it != numiter - thinning) {
											mean += dnorm(0, tau_u_chain[predind]);
										} else {
											mean += dnorm(0, tau_u);
										}
									}
									double missy = 0;
% if D == 'Normal':
									if (it != numiter - thinning) {
										missy = dnorm(mean, tau_chain[predind]);
									} else {
										missy = dnorm(mean, tau);
									}
% endif
% if D == 'Binomial':
% if link == 'logit':
									mean = exp(mean)/(1 + exp(mean));
% endif
% if link == 'probit':
									mean = norminv(mean);
% endif
% if link == 'cloglog':
									mean = 1.0 - exp(-exp(mean));
% endif
									missy = dbin(mean, ${popdata}_${n}[k]);
% endif
% if D == 'Poisson':
% if offset:
									mean += ${popdata}_${n}[i];
% endif
									missy = dpois(exp(mean));
% endif
% if D == '-ve Binomial':
% if offset:
									mean += ${popdata}_${n}[i];
% endif
									if (it != numiter - thinning) {
										missy = dpois(dgamma(alpha_chain[predind], alpha_chain[predind]) * exp(mean));
									} else {
										missy = dpois(dgamma(alpha, alpha) * exp(mean));
									}
/*
									if (it != numiter - thinning) { // FIXME: k refers to population dataset, whereas rho is length of sample dataset
										missy = dpois(rho_chain[(predind * mat_rho.rows()) + k] * exp(mean));
									} else {
										missy = dpois(rho[k] * exp(mean));
									}
*/
% endif
									//sumy += missy;
% if transform == 'Box-Cox':
% if lambtype == 'Modelled':
									double lambval = 0;
									double jacval = 0;
									if (it != numiter - thinning) {
										lambval = lamb_chain[predind];
										jacval = jac_chain[predind];
									} else {
										lambval = lamb;
										jacval = jac;
									}
									if (lambval * missy * jacval + 1.0 <= 0) { // Unable to log
										ys[id_k] = 0.0;
										zerocount[k]++;
										zerocountgrp[i]++;
										//std::cout << "Row: " << k << ", iteration: " << predind << " cannot be transformed (" << missy << "), setting to zero" << std::endl;
									} else {
										ys[id_k] = shift + std::exp(std::log1p(lambval*jacval*missy)/lambval);
									}
% else:
									if (lamb * missy * jac + 1.0 <= 0) { // Unable to log
										ys[id_k] = 0.0;
										zerocount[k]++;
										zerocountgrp[i]++;
										//std::cout << "Row: " << k << ", iteration: " << predind << " cannot be transformed (" << missy << "), setting to zero" << std::endl;
									} else {
										ys[id_k] = shift + std::exp(std::log1p(lamb*jac*missy)/lamb);
									}
% endif
% endif
% if transform == 'Log':
									ys[id_k] = shift + std::exp(missy);
% endif
% if transform == 'None':
									ys[id_k] = missy;
% endif
									if (!std::isfinite(ys[id_k])) {
										std::cout << "Transformation problem " << missy << " (row: " << id_k << ", area:" << i << ", urow:" << urow << ") does not give finite value." << std::endl;
										if (it != numiter - thinning) {
											for (size_t j = 0; j < mat_popx.cols(); ++j) {
												std::cout << "beta[" << j << "] = " << beta_chain[(predind * mat_beta.rows()) + j] << std::endl;
												std::cout << "x[" << j << "] = " << mat_popx(k, j) << std::endl;
											}
											if (urow != -1) {
												size_t uind = (predind * mat_u.rows()) + urow;
												std::cout << "u = " << u_chain[uind] << std::endl;
											} else {
												std::cout << "u unmatched" << std::endl;
											}
										} else { /* Values have not yet been written to the chain, so fetch from local variable */
											for (size_t j = 0; j < mat_popx.cols(); ++j) {
												std::cout << "beta[" << j << "] = " << mat_beta(j, 0) << std::endl;
												std::cout << "x[" << j << "] = " << mat_popx(k, j) << std::endl;
											}
											if (urow != -1) {
												std::cout << "u = " << mat_u(urow, 0) << std::endl;
											} else {
												std::cout << "u unmatched" << std::endl;
											}
										}
									}
									if (predind == 0) {
										y_M[k] = ys[id_k];
										y_S[k] = 0;
									} else {
										if (it != numiter - thinning) {
											double delta = ys[id_k] - y_M[k];
											y_M[k] += delta / (predind + 1);
											y_S[k] += delta * (ys[id_k] - y_M[k]);
										} else {
											y_pred[k] = ys[id_k];
										}
									}
								}
								double mean = std::accumulate(ys.cbegin(), ys.cend(), 0.0) / idmapsizes_${popdata}_${L2ID}[i];
								double sd = 0;
								for (const auto &val : ys) {
									sd += pow(val - mean, 2);
								}
								sd /= ys.size();
								sd = std::sqrt(sd);
								std::sort(ys.begin(), ys.end());
								auto yquant = [ys](auto q) {
									// Quantile function (in range 0-1), assumes that the vector is already sorted
									if (ys.size() == 1) {
										return ys[0];
									}
									double qpt = ((ys.size() - 1) * q) + 1;
									double qind = 0;
									double qfrac = std::modf(qpt, &qind);

									if (qind >= ys.size()) {
										qind = ys.size() - 1;
									}
									if (qind <= 0) {
										qind = 1;
									}
									return ys[qind - 1] * (1.0 - qfrac) + ys[qind] * qfrac;
								};

								double q20 = yquant(0.20);
								double q50 = yquant(0.50);
								double q80 = yquant(0.80);


% if povind:
								double headcount = 0;
								for (const auto &val : ys) {
									if (val < ${thresh}) {
										++headcount;
									}
								}
								headcount /= ys.size();

								double povgap = 0;
								for (const auto &val : ys) {
									if (val < ${thresh}) {
										povgap += (${thresh} - val) / ${thresh};
									}
								}
								povgap /= ys.size();


% endif
% if ineqind:
								double gini = 0;
								double sumy = 0;
								for (size_t i = 0; i < ys.size(); ++i) {
									gini += ys[i] * (i + 1);
									sumy += ys[i];
								}
								gini = 2 * gini / sumy - (ys.size() + 1);
								gini /= ys.size();

								double sumq20 = 0;
								double sumq80 = 0;
								for (const auto &val : ys) {
									if (val <= q20) {
										sumq20 += val;
									}
									if (val > q80) {
										sumq80 += val;
									}
								}
								double qshare = sumq80 / sumq20;
% endif
								if (it != numiter - thinning) {
									yj_mean_chain[(predind * mat_yj_mean.rows()) + i] = mean;
									yj_std_chain[(predind * mat_yj_mean.rows()) + i] = sd;
									yj_q10_chain[(predind * mat_yj_q10.rows()) + i] = yquant(0.10);
									yj_q20_chain[(predind * mat_yj_q20.rows()) + i] = q20;
									yj_q25_chain[(predind * mat_yj_q25.rows()) + i] = yquant(0.25);
									yj_q50_chain[(predind * mat_yj_q50.rows()) + i] = q50;
									yj_q75_chain[(predind * mat_yj_q75.rows()) + i] = yquant(0.75);
									yj_q80_chain[(predind * mat_yj_q80.rows()) + i] = q80;
									yj_q90_chain[(predind * mat_yj_q90.rows()) + i] = yquant(0.90);
% if povind:
									yj_headcount_chain[(predind * mat_yj_q90.rows()) + i] = headcount;
									yj_povgap_chain[(predind * mat_yj_q90.rows()) + i] = povgap;
% endif
% if ineqind:
									yj_gini_chain[(predind * mat_yj_q90.rows()) + i] = gini;
									yj_qshare_chain[(predind * mat_yj_q90.rows()) + i] = qshare;
% endif
								} else { /* Final iteration will overwrite chain value, so ovewrite source */
									mat_yj_mean(i, 0) = mean;
									mat_yj_std(i, 0) = sd;
									mat_yj_q10(i, 0) = yquant(0.10);
									mat_yj_q20(i, 0) = yquant(0.20);
									mat_yj_q25(i, 0) = yquant(0.25);
									mat_yj_q50(i, 0) = yquant(0.50);
									mat_yj_q75(i, 0) = yquant(0.75);
									mat_yj_q80(i, 0) = yquant(0.80);
									mat_yj_q90(i, 0) = yquant(0.90);
% if povind:
									mat_yj_headcount(i, 0) = headcount;
									mat_yj_povgap(i, 0) = povgap;
% endif
% if ineqind:
									mat_yj_gini(i, 0) = gini;
									mat_yj_qshare(i, 0) = qshare;
% endif
								}
							}
						}
					},
					p * chunksize,
					(p == ${numproc} - 1) ? p * chunksize + endchunksize : (p + 1) * chunksize
				);
			}
		}

		if (done == true) { /* Attempt to fix up parameters */
			std::cout << "Modelling finished, waiting to calculate predictions..." << std::endl;
			for (auto &val : procs) {
				val.wait();
			}

			auto &y_M = y_Ms[rngid];
			std::copy(y_M.cbegin(), y_M.cend(), y_pred_M);
			auto &y_S = y_Ss[rngid];
			std::copy(y_S.cbegin(), y_S.cend(), y_pred_S);

			auto &zerocount = zerocounts[rngid];
			auto &zerocountgrp = zerocountsgrp[rngid];
			size_t totzero = std::accumulate(zerocount.cbegin(), zerocount.cend(), 0.0);
			if (totzero > 0) {
				std::cout << "The following rows were reset to zero: \\n";
				std::cout << "row\tcount\\n";
				for (size_t i = 0; i < zerocount.size(); ++i) {
					if (zerocount[i] > 0) {
						std::cout << (i+1) << "\t" << zerocount[i] << "\\n";
					}
				}
				std::cout << "avg\t" << static_cast<double>(totzero) / zerocount.size() << "\\n";

				std::cout << "The following groups were reset to zero: \\n";
				std::cout << "group\tcount\\n";
				for (size_t i = 0; i < zerocountgrp.size(); ++i) {
					if (zerocountgrp[i] > 0) {
						std::cout << (i+1) << "\t" << zerocountgrp[i] << "\\n";
					}
				}
				std::cout << "avg\t" << static_cast<double>(totzero) / zerocountgrp.size() << "\\n";

				std::cout << "total\t" << totzero << std::endl;
			}
		}
	}
'''

    diccode = '''
% if D == 'Normal':
                RectMatrix ydiff(mat_y);
                for (int i = 0; i < ydiff.rows(); i++) {
                    ydiff(i, 0) -= double(u[int(${L2ID}[i])]);
                }
                RectMatrix yxb = ydiff - mat_x * mat_beta;
                deviance = (tau*(yxb.T() * yxb)(0,0)) + (log(3.14159265)-log(tau))*length(${y}) + (2.0*0.346573590279973*length(${y})); // Last term is log(sqrt(2))
% else:
                deviance = 0.0;
                for (int j = 0; j < mat_x.rows(); ++j) {
                    double xb = 0;
                    for (int k = 0; k < mat_x.cols(); ++k) {
                        xb += mat_x(j, k) * mat_beta(k, 0);
                    }
                    xb += mat_u(int(${L2ID}[j]), 0);
% if D == 'Binomial':
% if link == 'logit':
                    double temp = -log1p(exp(-xb));
                    deviance += (mat_n(j, 0) - mat_y(j, 0)) * (-xb + temp) + mat_y(j, 0) * temp;
% endif
% if link == 'probit':
                    double temp = normprob(-xb);
                    deviance += (mat_n(j, 0) - mat_y(j, 0)) * log(temp) + mat_y(j, 0) * log(1.0 - temp);
% endif
% if link == 'cloglog':
                    double temp = -exp(xb);
                    deviance += (mat_n(j, 0) - mat_y(j, 0)) * temp + mat_y(j, 0) * log(1.0 - exp(temp));
% endif
                    deviance += logfact(mat_n(j, 0)) - logfact(mat_y(j, 0)) - logfact(mat_n(j, 0) - mat_y(j, 0));
% endif
% if D == 'Poisson':
% if offset:
                    xb += mat_n(j, 0);
% endif
                    deviance += mat_y(j, 0) * xb - exp(xb);
                    deviance += logfact(mat_y(j, 0));
% endif
% if D == '-ve Binomial':
% if offset:
                    xb += mat_n(j, 0);
% endif
                    deviance += mat_y(j, 0) * xb - exp(xb) * mat_rho(j, 0);
                    deviance -= logfact(mat_y(j, 0));
% endif
                }
                deviance *= -2;
% endif
'''


    preparedata = '''
import numpy
import scipy
import scipy.stats

mydata = data['datafile']
mydata2 = data[str(popdata)]

missid = mydata2.variables[L2ID]['data'].mask
for var in x.orignames:
    missid = missid | mydata2.variables[var]['data'].mask
nummiss = numpy.sum(missid)

shortid = mydata.variables[L2ID]['data']
longid = mydata2.variables[popid]['data']

uniqshort = numpy.unique(shortid)
lookup = {key: value for (key, value) in zip(uniqshort, range(0, len(uniqshort)))}

idmatch = numpy.ma.zeros(mydata2.nobs)
for i in range(0, len(idmatch)):
    idmatch[i] = lookup.get(longid[i], -1)
mydata2.addvariable("_idmatch", data = idmatch)

if povind:
    if userthresh == 'Automatic':
        context['thresh'].set_value(str(0.6 * numpy.median(mydata.variables[y]['data'])))

def geomean(x):
    return numpy.exp(numpy.log(x).mean())

def bc(y, lamb, jac):
    return (y**lamb - 1) / (lamb * jac)

if D == 'Normal':
    yout = mydata.variables[y]['data']
    ymin = yout.min()
    shift = 0.0
    if ymin <= 0:
        shift = ymin - 1.0

    shiftval._name = str(shift)

    if transform == 'Box-Cox' or transform == 'Log':
        yout = yout - shift
        context['shift'].set_value(shift)

    if transform == 'Box-Cox':
        if lambtype == 'Automatic':
            (yout, lamb) = scipy.stats.boxcox(yout)
            lambval._name = str(lamb)
            context['lamb'].set_value(lamb)
        if lambtype == 'Manual':
            lamb = float(lambval)
            jac = geomean(yout)**(lamb - 1)
            #yout = scipy.stats.boxcox(yout, lamb)
            yout = bc(yout, lamb, jac)
            context['lamb'].set_value(lamb)
            context['jac'].set_value(jac)

    if transform == 'Log':
        yout = numpy.log(yout)

    mydata.variables[y]['data'] = yout
'''